@@ -38,8 +38,8 @@ set *
 add_set(const char *name, set *q)
 {
 	int pos;
-	elem *ll = xmalloc(sizeof(*ll));
-	elem *w;
+	set_elem *ll = xmalloc(sizeof(*ll));
+	set_elem *w;
 
 	if (q == NULL)
 		q = create_set();
@@ -68,8 +68,8 @@ add_set_unique(const char *name, set *q, bool *unique)
 {
 	unsigned int hash;
 	int pos;
-	elem *ll;
-	elem *w;
+	set_elem *ll;
+	set_elem *w;
 	bool uniq = false;
 
 	if (q == NULL)
@@ -117,8 +117,8 @@ add_set_value(const char *name, void *ptr, set *q)
 {
 	unsigned int hash;
 	int pos;
-	elem *ll;
-	elem *w;
+	set_elem *ll;
+	set_elem *w;
 
 	hash = fnv1a32(name);
 	pos = hash % _SET_HASH_SIZE;
@@ -154,7 +154,7 @@ contains_set(const char *name, set *q)
 {
 	unsigned int hash;
 	int pos;
-	elem *w;
+	set_elem *w;
 	bool found;
 
 	hash = fnv1a32(name);
@@ -180,7 +180,7 @@ get_set(const char *name, set *q)
 {
 	unsigned int hash;
 	int pos;
-	elem *w;
+	set_elem *w;
 
 	hash = fnv1a32(name);
 	pos = hash % _SET_HASH_SIZE;
@@ -195,19 +195,26 @@ get_set(const char *name, set *q)
 	return NULL;
 }
 
-/* remove elem from a set. matches ->name and frees name,item */
-set *
+/* remove elem from a set. matches ->name and frees name,item, returns
+ * val if removed, NULL otherwise
+ * note that when val isn't set, NULL is returned, so the caller should
+ * use the removed argument to determine if something was removed from
+ * the set. */
+void *
 del_set(const char *s, set *q, bool *removed)
 {
 	unsigned int hash;
 	int pos;
-	elem *ll;
-	elem *w;
+	set_elem *ll;
+	set_elem *w;
+	void *ret;
+	bool rmd;
 
 	hash = fnv1a32(s);
 	pos = hash % _SET_HASH_SIZE;
 
-	*removed = false;
+	ret = NULL;
+	rmd = false;
 	if (q->buckets[pos] != NULL) {
 		ll = NULL;
 		for (w = q->buckets[pos]; w != NULL; ll = w, w = w->next) {
@@ -217,17 +224,20 @@ del_set(const char *s, set *q, bool *removed)
 				} else {
 					ll->next = w->next;
 				}
+				ret = w->val;
 				free(w->name);
 				free(w);
-				*removed = true;
+				rmd = true;
 				break;
 			}
 		}
 	}
 
-	if (*removed)
+	if (rmd)
 		q->len--;
-	return q;
+	if (removed != NULL)
+		*removed = rmd;
+	return ret;
 }
 
 /* return the contents of a set as an array of strings
@@ -238,7 +248,7 @@ size_t
 list_set(set *q, char ***l)
 {
 	int i;
-	elem *w;
+	set_elem *w;
 	char **ret;
 
 	ret = *l = xmalloc(sizeof(char **) * (q->len + 1));
@@ -252,11 +262,27 @@ list_set(set *q, char ***l)
 	return q->len;
 }
 
+size_t
+array_set(set *q, array_t *ret)
+{
+	int i;
+	set_elem *w;
+	array_t blank = array_init_decl;
+
+	*ret = blank;
+	for (i = 0; i < _SET_HASH_SIZE; i++) {
+		for (w = q->buckets[i]; w != NULL; w = w->next)
+			xarraypush_ptr(ret, w->name);
+	}
+
+	return q->len;
+}
+
 size_t
 values_set(set *q, array_t *ret)
 {
 	int i;
-	elem *w;
+	set_elem *w;
 	array_t blank = array_init_decl;
 
 	*ret = blank;
@@ -279,8 +305,8 @@ void
 clear_set(set *q)
 {
 	int i;
-	elem *w;
-	elem *e;
+	set_elem *w;
+	set_elem *e;
 
 	for (i = 0; i < _SET_HASH_SIZE; i++) {
 		for (w = q->buckets[i]; w != NULL; w = e) {
@@ -305,7 +331,7 @@ free_set(set *q)
 static void
 print_set(const set *q)
 {
-	elem *w;
+	set_elem *w;
 	int i;
 
 	for (i = 0; i < _SET_HASH_SIZE; i++) {
