@@ -12,19 +12,19 @@
 
 #include "xarray.h"
 
-typedef struct elem_t elem;
+typedef struct elem_t set_elem;
 typedef struct set_t set;
 
 struct elem_t {
 	char *name;
 	unsigned int hash;  /* FNV1a32 */
 	void *val;
-	elem *next;
+	set_elem *next;
 };
 
 #define _SET_HASH_SIZE 128
 struct set_t {
-	elem *buckets[_SET_HASH_SIZE];
+	set_elem *buckets[_SET_HASH_SIZE];
 	size_t len;
 };
 
@@ -34,8 +34,9 @@ set *add_set_unique(const char *name, set *q, bool *unique);
 void *add_set_value(const char *name, void *ptr, set *q);
 bool contains_set(const char *name, set *q);
 void *get_set(const char *name, set *q);
-set *del_set(const char *s, set *q, bool *removed);
+void *del_set(const char *s, set *q, bool *removed);
 size_t list_set(set *q, char ***l);
+size_t array_set(set *q, array_t *ret);
 size_t values_set(set *q, array_t *ret);
 size_t cnt_set(set *q);
 void free_set(set *q);
