@@ -22,6 +22,7 @@
 #include "eat_file.h"
 #include "rmspace.h"
 #include "scandirat.h"
+#include "set.h"
 #include "xasprintf.h"
 
 /* variables to control runtime behavior */
@@ -31,6 +32,7 @@ int verbose = 0;
 int quiet = 0;
 int twidth;
 char pretend = 0;
+char *portarch;
 char *portroot;
 char *config_protect;
 char *config_protect_mask;
@@ -46,7 +48,6 @@ DECLARE_ARRAY(overlays);
 DECLARE_ARRAY(overlay_names);
 DECLARE_ARRAY(overlay_src);
 
-static char *portarch;
 static char *portedb;
 static char *eprefix;
 static char *accept_license;
@@ -224,8 +225,6 @@ makeargv(const char *string, int *argc, char ***argv)
 
 	*argc = 1;
 	(*argv)[0] = xstrdup(argv0);
-	q = xstrdup(string);
-	str = q;
 
 	/* shortcut empty strings */
 	while (isspace((int)*string))
@@ -233,6 +232,9 @@ makeargv(const char *string, int *argc, char ***argv)
 	if (*string == '\0')
 		return;
 
+	q = xstrdup(string);
+	str = q;
+
 	remove_extra_space(str);
 	rmspace(str);
 
@@ -353,10 +355,12 @@ set_portage_env_var(env_vars *var, const char *value, const char *src)
 	}
 }
 
-/* Helper to read a portage env file (e.g. make.conf), or recursively if
- * it points to a directory */
+/* Helper to read a portage file (e.g. make.conf, package.mask), or
+ * recursively if it points to a directory (we don't care about EAPI for
+ * dirs, basically PMS 5.2.5 EAPI restriction is ignored) */
+enum portage_file_type { ENV_FILE, PMASK_FILE };
 static void
-read_portage_env_file(const char *file, env_vars vars[])
+read_portage_file(const char *file, enum portage_file_type type, void *data)
 {
 	FILE *fp;
 	struct dirent **dents;
@@ -367,6 +371,8 @@ read_portage_env_file(const char *file, env_vars vars[])
 	size_t buflen = 0;
 	size_t line;
 	int i;
+	env_vars *vars = data;
+	set *masks = data;
 
 	if (getenv("DEBUG"))
 		fprintf(stderr, "profile %s\n", file);
@@ -383,7 +389,7 @@ read_portage_env_file(const char *file, env_vars vars[])
 					d->d_name[strlen(d->d_name) - 1] == '~')
 				continue;
 			snprintf(npath, sizeof(npath), "%s/%s", file, d->d_name);
-			read_portage_env_file(npath, vars);
+			read_portage_file(npath, type, data);
 		}
 		scandir_free(dents, dentslen);
 		goto done;
@@ -401,84 +407,107 @@ read_portage_env_file(const char *file, env_vars vars[])
 			continue;
 
 		/* Handle "source" keyword */
-		if (strncmp(buf, "source ", 7) == 0) {
-			const char *sfile = buf + 7;
-			char npath[_Q_PATH_MAX * 2];
+		if (type == ENV_FILE) {
+			if (strncmp(buf, "source ", 7) == 0) {
+				const char *sfile = buf + 7;
+				char npath[_Q_PATH_MAX * 2];
 
-			if (sfile[0] != '/') {
-				/* handle relative paths */
-				size_t file_path_len;
+				if (sfile[0] != '/') {
+					/* handle relative paths */
+					size_t file_path_len;
 
-				s = strrchr(file, '/');
-				file_path_len = s - file + 1;
+					s = strrchr(file, '/');
+					file_path_len = s - file + 1;
 
-				snprintf(npath, sizeof(npath), "%.*s/%s",
-						(int)file_path_len, file, sfile);
-				sfile = npath;
-			}
-
-			read_portage_env_file(sfile, vars);
-			continue;
-		}
-
-		/* look for our desired variables and grab their value */
-		for (i = 0; vars[i].name; ++i) {
-			if (buf[vars[i].name_len] != '=' && buf[vars[i].name_len] != ' ')
-				continue;
-			if (strncmp(buf, vars[i].name, vars[i].name_len))
-				continue;
-
-			/* make sure we handle spaces between the varname, the =,
-			 * and the value:
-			 * VAR=val   VAR = val   VAR="val"
-			 */
-			s = buf + vars[i].name_len;
-			if ((p = strchr(s, '=')) != NULL)
-				s = p + 1;
-			while (isspace(*s))
-				++s;
-			if (*s == '"' || *s == '\'') {
-				char *endq;
-				char q = *s;
-
-				/* make sure we handle spacing/comments after the quote */
-				endq = strchr(s + 1, q);
-				if (!endq) {
-					/* If the last char is not a quote, then we span lines */
-					size_t abuflen;
-					char *abuf;
-
-					abuf = NULL;
-					while (getline(&abuf, &abuflen, fp) != -1) {
-						buf = xrealloc(buf, buflen + abuflen);
-						endq = strchr(abuf, q);
-						if (endq)
-							*endq = '\0';
-
-						strcat(buf, abuf);
-						buflen += abuflen;
-
-						if (endq)
-							break;
-					}
-					free(abuf);
-
-					if (!endq)
-						warn("%s:%zu: %s: quote mismatch",
-								file, line, vars[i].name);
-
-					s = buf + vars[i].name_len + 2;
-				} else {
-					*endq = '\0';
-					++s;
+					snprintf(npath, sizeof(npath), "%.*s/%s",
+							(int)file_path_len, file, sfile);
+					sfile = npath;
 				}
-			} else {
-				/* no quotes, so chop the spacing/comments ourselves */
-				size_t off = strcspn(s, "# \t\n");
-				s[off] = '\0';
+
+				read_portage_file(sfile, type, data);
+				continue;
 			}
 
-			set_portage_env_var(&vars[i], s, file);
+			/* look for our desired variables and grab their value */
+			for (i = 0; vars[i].name; i++) {
+				if (buf[vars[i].name_len] != '=' &&
+						buf[vars[i].name_len] != ' ')
+					continue;
+				if (strncmp(buf, vars[i].name, vars[i].name_len))
+					continue;
+
+				/* make sure we handle spaces between the varname, the =,
+				 * and the value:
+				 * VAR=val   VAR = val   VAR="val"
+				 */
+				s = buf + vars[i].name_len;
+				if ((p = strchr(s, '=')) != NULL)
+					s = p + 1;
+				while (isspace(*s))
+					s++;
+				if (*s == '"' || *s == '\'') {
+					char *endq;
+					char q = *s;
+
+					/* make sure we handle spacing/comments after the quote */
+					endq = strchr(s + 1, q);
+					if (!endq) {
+						/* if the last char is not a quote,
+						 * then we span lines */
+						size_t abuflen;
+						char *abuf;
+
+						abuf = NULL;
+						while (getline(&abuf, &abuflen, fp) != -1) {
+							buf = xrealloc(buf, buflen + abuflen);
+							endq = strchr(abuf, q);
+							if (endq)
+								*endq = '\0';
+
+							strcat(buf, abuf);
+							buflen += abuflen;
+
+							if (endq)
+								break;
+						}
+						free(abuf);
+
+						if (!endq)
+							warn("%s:%zu: %s: quote mismatch",
+									file, line, vars[i].name);
+
+						s = buf + vars[i].name_len + 2;
+					} else {
+						*endq = '\0';
+						s++;
+					}
+				} else {
+					/* no quotes, so chop the spacing/comments ourselves */
+					size_t off = strcspn(s, "# \t\n");
+					s[off] = '\0';
+				}
+
+				set_portage_env_var(&vars[i], s, file);
+			}
+		} else if (type == PMASK_FILE) {
+			/* trim leading space */
+			for (s = buf; isspace((int)*s); s++)
+				;
+			if (*s == '\0')
+				continue;
+			if (*s == '-') {
+				/* negation/removal, lookup and drop mask if it exists;
+				 * note that this only supports exact matches (PMS
+				 * 5.2.5) so we don't even have to parse and use
+				 * atom-compare here */
+				s++;
+				if ((p = del_set(s, masks, NULL)) != NULL)
+					free(p);
+			} else {
+				p = xstrdup(file);
+				if (add_set_value(s, p, masks) != NULL)
+					free(p);
+			}
 		}
 	}
 
@@ -489,9 +518,10 @@ read_portage_env_file(const char *file, env_vars vars[])
 
 /* Helper to recursively read stacked make.defaults in profiles */
 static void
-read_portage_profile(const char *profile, env_vars vars[])
+read_portage_profile(const char *profile, env_vars vars[], set *masks)
 {
 	char profile_file[_Q_PATH_MAX * 3];
+	char rpath[_Q_PATH_MAX];
 	size_t profile_len;
 	char *s;
 	char *p;
@@ -510,8 +540,11 @@ read_portage_profile(const char *profile, env_vars vars[])
 	 * treat parent profiles as defaults, that can be overridden by
 	 * *this* profile. */
 	strcpy(profile_file + profile_len, "parent");
-	if (eat_file(profile_file, &buf, &buf_len) == 0)
+	if (eat_file(profile_file, &buf, &buf_len) == 0) {
+		if (buf != NULL)
+			free(buf);
 		return;
+	}
 
 	s = strtok_r(buf, "\n", &saveptr);
 	while (s) {
@@ -544,7 +577,9 @@ read_portage_profile(const char *profile, env_vars vars[])
 			snprintf(profile_file + profile_len,
 					sizeof(profile_file) - profile_len, "%s", s);
 		}
-		read_portage_profile(profile_file, vars);
+		read_portage_profile(
+				realpath(profile_file, rpath) == NULL ? profile_file : rpath,
+				vars, masks);
 		/* restore original path in case we were repointed by profile */
 		if (p != NULL)
 			snprintf(profile_file, sizeof(profile_file), "%s/", profile);
@@ -553,9 +588,11 @@ read_portage_profile(const char *profile, env_vars vars[])
 
 	free(buf);
 
-	/* now consume *this* profile's make.defaults */
+	/* now consume *this* profile's make.defaults and package.mask */
 	strcpy(profile_file + profile_len, "make.defaults");
-	read_portage_env_file(profile_file, vars);
+	read_portage_file(profile_file, ENV_FILE, vars);
+	strcpy(profile_file + profile_len, "package.mask");
+	read_portage_file(profile_file, PMASK_FILE, masks);
 }
 
 static bool nocolor = 0;
@@ -594,14 +631,18 @@ env_vars vars_to_read[] = {
 
 #undef _Q_EV
 };
+set *package_masks = NULL;
 
 /* Handle a single file in the repos.conf format. */
 static void
-read_one_repos_conf(const char *repos_conf)
+read_one_repos_conf(const char *repos_conf, char **primary)
 {
 	int nsec;
 	char *conf;
-	const char *main_repo, *repo, *path;
+	char rrepo[_Q_PATH_MAX];
+	const char *main_repo;
+	const char *repo;
+	const char *path;
 	dictionary *dict;
 
 	if (getenv("DEBUG"))
@@ -614,7 +655,7 @@ read_one_repos_conf(const char *repos_conf)
 	nsec = iniparser_getnsec(dict);
 	while (nsec-- > 0) {
 		repo = iniparser_getsecname(dict, nsec);
-		if (!strcmp(repo, "DEFAULT"))
+		if (strcmp(repo, "DEFAULT") == 0)  /* already handled above */
 			continue;
 
 		xasprintf(&conf, "%s:location", repo);
@@ -624,6 +665,10 @@ read_one_repos_conf(const char *repos_conf)
 			size_t n;
 			char *overlay;
 
+			/* try not to get confused by symlinks etc. */
+			if (realpath(path, rrepo) != NULL)
+				path = rrepo;
+
 			array_for_each(overlay_names, n, overlay) {
 				if (strcmp(overlay, repo) == 0)
 					break;
@@ -639,14 +684,11 @@ read_one_repos_conf(const char *repos_conf)
 				ele = array_get_elem(overlays, n) = xstrdup(path);
 			} else {
 				ele = xarraypush_str(overlays, path);
-				xarraypush_str(overlay_names, repo);
+				overlay = xarraypush_str(overlay_names, repo);
 				xarraypush_str(overlay_src, repos_conf);
 			}
-			if (main_repo && strcmp(repo, main_repo) == 0) {
-				main_overlay = ele;
-				free(vars_to_read[11 /* PORTDIR */].src);
-				vars_to_read[11 /* PORTDIR */].src = xstrdup(repos_conf);
-			}
+			if (main_repo && strcmp(repo, main_repo) == 0)
+				*primary = overlay;
 		}
 		free(conf);
 	}
@@ -656,7 +698,7 @@ read_one_repos_conf(const char *repos_conf)
 
 /* Handle a possible directory of files. */
 static void
-read_repos_conf(const char *configroot, const char *repos_conf)
+read_repos_conf(const char *configroot, const char *repos_conf, char **primary)
 {
 	char *top_conf, *sub_conf;
 	int i, count;
@@ -668,7 +710,7 @@ read_repos_conf(const char *configroot, const char *repos_conf)
 	count = scandir(top_conf, &confs, NULL, alphasort);
 	if (count == -1) {
 		if (errno == ENOTDIR)
-			read_one_repos_conf(top_conf);
+			read_one_repos_conf(top_conf, primary);
 	} else {
 		for (i = 0; i < count; ++i) {
 			const char *name = confs[i]->d_name;
@@ -700,7 +742,7 @@ read_repos_conf(const char *configroot, const char *repos_conf)
 				}
 			}
 
-			read_one_repos_conf(sub_conf);
+			read_one_repos_conf(sub_conf, primary);
 			free(sub_conf);
 		}
 		scandir_free(confs, count);
@@ -715,10 +757,11 @@ initialize_portage_env(void)
 	const char *s;
 	env_vars *var;
 	char pathbuf[_Q_PATH_MAX];
+	char rpathbuf[_Q_PATH_MAX];
 	const char *configroot = getenv("PORTAGE_CONFIGROOT");
-	char *orig_main_overlay = main_overlay;
+	char *primary_overlay = NULL;
 
-	/* initialize all the strings with their default value */
+	/* initialize all the properties with their default value */
 	for (i = 0; vars_to_read[i].name; ++i) {
 		var = &vars_to_read[i];
 		if (var->type != _Q_BOOL)
@@ -726,6 +769,8 @@ initialize_portage_env(void)
 		var->src = xstrdup(STR_DEFAULT);
 	}
 
+	package_masks = create_set();
+
 	/* figure out where to find our config files */
 	if (!configroot)
 		configroot = CONFIG_EPREFIX;
@@ -736,47 +781,54 @@ initialize_portage_env(void)
 		i--;
 
 	/* read overlays first so we can resolve repo references in profile
-	 * parent files */
+	 * parent files (non PMS feature?) */
 	snprintf(pathbuf, sizeof(pathbuf), "%.*s", (int)i, configroot);
-	read_repos_conf(pathbuf, "/usr/share/portage/config/repos.conf");
-	read_repos_conf(pathbuf, "/etc/portage/repos.conf");
-	if (orig_main_overlay != main_overlay)
-		free(orig_main_overlay);
-	if (array_cnt(overlays) == 0) {
-		xarraypush_ptr(overlays, main_overlay);
-		xarraypush_str(overlay_names, "<PORTDIR>");
-		xarraypush_str(overlay_src, STR_DEFAULT);
-	} else if (orig_main_overlay == main_overlay) {
-		/* if no explicit overlay was flagged as main, take the first one */
-		main_overlay = array_get_elem(overlays, 0);
-		set_portage_env_var(&vars_to_read[11] /* PORTDIR */, main_overlay,
-				(char *)array_get_elem(overlay_src, 0));
-	}
+	read_repos_conf(pathbuf, "/usr/share/portage/config/repos.conf",
+			&primary_overlay);
+	read_repos_conf(pathbuf, "/etc/portage/repos.conf", &primary_overlay);
 
 	/* consider Portage's defaults */
 	snprintf(pathbuf, sizeof(pathbuf),
 			"%.*s/usr/share/portage/config/make.globals",
 			(int)i, configroot);
-	read_portage_env_file(pathbuf, vars_to_read);
+	read_portage_file(pathbuf, ENV_FILE, vars_to_read);
+
+	/* start with base masks, Portage behaviour PMS 5.2.8 */
+	if (primary_overlay != NULL) {
+		char *overlay;
+		size_t n;
+		array_for_each(overlay_names, n, overlay) {
+			if (overlay == primary_overlay) {
+				snprintf(pathbuf, sizeof(pathbuf), "%s/profiles/package.mask",
+						(char *)array_get_elem(overlays, n));
+				read_portage_file(pathbuf, PMASK_FILE, package_masks);
+				break;
+			}
+		}
+	}
 
 	/* walk all the stacked profiles */
 	snprintf(pathbuf, sizeof(pathbuf), "%.*s/etc/make.profile",
 			(int)i, configroot);
-	read_portage_profile(pathbuf, vars_to_read);
+	read_portage_profile(
+			realpath(pathbuf, rpathbuf) == NULL ? pathbuf : rpathbuf,
+			vars_to_read, package_masks);
 	snprintf(pathbuf, sizeof(pathbuf), "%.*s/etc/portage/make.profile",
 			(int)i, configroot);
-	read_portage_profile(pathbuf, vars_to_read);
+	read_portage_profile(
+			realpath(pathbuf, rpathbuf) == NULL ? pathbuf : rpathbuf,
+			vars_to_read, package_masks);
 
-	/* now read all the config files */
+	/* now read all Portage's config files */
 	snprintf(pathbuf, sizeof(pathbuf), "%.*s/etc/make.conf",
 			(int)i, configroot);
-	read_portage_env_file(pathbuf, vars_to_read);
+	read_portage_file(pathbuf, ENV_FILE, vars_to_read);
 	snprintf(pathbuf, sizeof(pathbuf), "%.*s/etc/portage/make.conf",
 			(int)i, configroot);
-	read_portage_env_file(pathbuf, vars_to_read);
+	read_portage_file(pathbuf, ENV_FILE, vars_to_read);
 
 	/* finally, check the env */
-	for (i = 0; vars_to_read[i].name; ++i) {
+	for (i = 0; vars_to_read[i].name; i++) {
 		var = &vars_to_read[i];
 		s = getenv(var->name);
 		if (s != NULL)
@@ -854,6 +906,57 @@ initialize_portage_env(void)
 		}
 	}
 
+	/* handle PORTDIR and primary_overlay to get a unified
+	 * administration in overlays */
+	{
+		char *overlay;
+		var = &vars_to_read[11];  /* PORTDIR */
+
+		if (strcmp(var->src, STR_DEFAULT) != 0 || array_cnt(overlays) == 0) {
+			char roverlay[_Q_PATH_MAX];
+			/* get cannonical path, we do so for repos.conf too */
+			if (realpath(main_overlay, roverlay) == NULL)
+				snprintf(roverlay, sizeof(roverlay), "%s", main_overlay);
+			array_for_each(overlays, i, overlay) {
+				if (strcmp(overlay, roverlay) == 0)
+					break;
+				overlay = NULL;
+			}
+			if (overlay == NULL) {  /* add PORTDIR to overlays */
+				free(main_overlay);
+				main_overlay = xstrdup(roverlay);
+				xarraypush_ptr(overlays, main_overlay);
+				xarraypush_str(overlay_names, "<PORTDIR>");
+				xarraypush_str(overlay_src, var->src);
+			} else {
+				/* ignore make.conf and/or env setting origin if defined by
+				 * repos.conf since the former are deprecated */
+				free(main_overlay);
+			}
+			main_overlay = NULL;  /* now added to overlays */
+		} else {
+			free(main_overlay);
+		}
+
+		/* set main_overlay to the one pointed to by repos.conf, if any */
+		i = 0;
+		if (primary_overlay != NULL) {
+			array_for_each(overlay_names, i, overlay) {
+				if (overlay == primary_overlay)
+					break;
+				overlay = NULL;
+			}
+			/* if no explicit overlay was flagged as main, take the
+			 * first one */
+			if (overlay == NULL)
+				i = 0;
+		}
+		main_overlay = array_get_elem(overlays, i);
+		/* set source for PORTDIR var */
+		free(var->src);
+		var->src = xstrdup((char *)array_get_elem(overlay_src, i));
+	}
+
 	/* Make sure ROOT always ends in a slash */
 	var = &vars_to_read[0];
 	if (var->value_len == 0 || (*var->value.s)[var->value_len - 1] != '/') {
